from numpy import zeros, array
from numpy.linalg import det, inv, pinv
from math import floor
from itertools import product
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 21 17:31:43 2020

auxiliary functions for the Trendfitter library

- Score calculation with missing values

@author: Daniel Rodrigues
"""

def scores_with_missing_values( omega, loadings, X_matrix, LVs = None, method = 'TSM', inverse_method = None, inverse_term = None ):
    """

    function to estimate missing values using different techniques. 
    
    TSR - Trimmed Score Regression  - As in Arteaga & Ferrer's " Dealing with missing data in MSPC: several methods, different interpretations, some examples" (2002)

    CMR - Conditional Mean Replacement - as in Nelston, et al.'s "Missing Data methods in PCA and PLS: Score calculations with incomplete observations" (1996)

    Parameters:

    Omega

    loadings

    X_matrix

    method:

    inverse_method:

    inverse_term:
    
    returns
    
    score: vector of predicted scores

    """
    if LVs == None : LVs = omega.shape[0]

    if method == 'TSR' : #estimate using the 'TSR' method - Trimmed Score Regression
        
        B_1 = omega[ :LVs, :LVs ] @ loadings[ :LVs ] @ loadings[ :LVs ].T  # OMEGA.P*'.P*
        B_2 = pinv( loadings[ :LVs ] @ loadings.T @ omega @ loadings @ loadings[ :LVs ].T ) # (P*'.P*.OMEGA.P'.P*)^-1 
        B_factor = B_1 @ B_2 @ loadings[ :LVs ]   # OMEGA.P*'.P*.(P*'.P*.OMEGA.P.P*')^-1.P*'
       # print(B_factor)
        scores = ( B_factor @ X_matrix.T ).reshape( X_matrix.shape[ 0 ], -1 )         

    elif method == 'CMR' : # estimate using the 'CMR' method - Conditional mean replacement method
        B_1 = omega[ :LVs, :LVs ] @ loadings[ :LVs ] # OMEGA.P*'
        B_2 = pinv( loadings.T @ omega @ loadings )  #(P*.OMEGA.P*')^-1
        B_factor = B_1 @ B_2 #OMEGA.P*'.(P*.OMEGA.P*')^-1
        #print(B_factor)
        scores = ( B_factor @ X_matrix.T ).reshape( X_matrix.shape[ 0 ], -1 )
    
    elif method == 'TSM' : # Trimmed score method - just substitute all nans for zeros
        scores = X_matrix @ loadings[ :LVs ].T

    elif method == 'PMP' : #Projection to the Model Plane method
        B_factor = pinv( loadings[ :LVs ] @ loadings[ :LVs ].T ) @ loadings[ :LVs ]
        #print(B_factor)
        scores = ( B_factor @ X_matrix.T ).reshape( X_matrix.shape[ 0 ], -1 )

    else : raise Exception('User inputted method "{}" not implemented'.format(method))

    return scores

    """
def subinverse( main_inverse_term, desired_variables ):
        
    
    Because of the necessity of calculating multiple matrix inverses, this functions uses a method 
    published in Juarez-Ruiz, E. et al.'s "Relationship between the Inverses of a Matrix and a Submatrix"
    to calculate such inverses using a method that uses the full-inverse as basis. (2016) 
    
    
    final_inv_num_total = zeros( main_inverse_term.shape[ 0 ] - len( desired_variables ) )
    i, j = 0, 0
    final_inv_den_matrix = zeros( len( desired_variables ) )
    
    i = 0
    for q, p in product( desired_variables, repeat = 2 ):
        final_inv_den_matrix[ i % len( desired_variables ) , floor( i / len ( desired_variables ) ) ] = main_inverse_term[ q, p ]
        i += 1

    final_inv_den = det( final_inv_den_matrix )
    
    for i in range( main_inverse_term.shape[ 0 ] - len( desired_variables ) ):
        
        for j in range( main_inverse_term.shape[ 0 ] - len( desired_variables ) ):
            
            final_inv_num = zeros( main_inverse_term.shape[ 0 ] - len( desired_variables ) )
            final_inv_num[ :-1, :-1 ] = final_inv_den_matrix
            
            k = 0
            for qp in desired_variables:
                final_inv_num[ -1, k ] = main_inverse_term[ i, qp ]
                final_inv_num[ k, -1 ] = main_inverse_term[ qp, j ]
                k += 1
            final_inv_num[ -1, -1 ] = main_inverse_term[ i, j ]
            
            final_inv_num_total[ i , j ] = final_inv_num
            
            j += 1
        
        i += 1
        
        inverse_matrix = det( final_inv_num_total ) / final_inv_den
        
    return inverse_matrix
"""


from .PCA import PCA
from .PLS import PLS
from .MB_PCA import MB_PCA
from .MB_PLS import MB_PLS
from .SMB_PLS import SMB_PLS
from .MLSMB_PLS import MLSMB_PLS
from .DiPLS import DiPLS

__all__ = ['PCA', 'PLS', 'MB_PCA', 'MB_PLS', 'SMB_PLS', 'MLSMB_PLS', 'DiPLS']



from numpy import array, isnan, nansum, nan_to_num, multiply, sum, sqrt, \
                  append, zeros, place, nan, concatenate, mean, nanvar,  \
                  std, unique, where, nanmean, diagonal
from numpy.linalg import norm, pinv
from sklearn.model_selection import KFold
from pandas import DataFrame, Series
from trendfitter.auxiliary.tf_aux import scores_with_missing_values
from . import PLS

class MB_PLS:
    """
    A sklearn-like class for the Multi-Block Projection to Latent Structures.

        Parameters
        ----------
        cv_splits_number : int, optional
            number of splits used for cross validation in case latent_variables is None
        tol : float, optional
            value used to decide if model has converged
        loop_limit : int, optional
            maximum number of loops for the extraction of one latent variable
        missing_values_method : str, optional
            string to define the method that the model will deal with missing values

        Attributes
        ----------
        latent_variables : int
            list of number of latent variables deemed relevant from each block. 
        block_divs : [int]
            list with the index of every block final position. ex. [2,4] means two blocks with 
                columns 0 and 1 on the first block and columns 2 and 3 on the second block. 
                Assigned when fit method runs.
        p_loadings_block : [array_like]
            list of p loadings arrays of every block with all the extracted latent variables 
        superlevel_p_loadings : array_like
            array of all latent variables extracted p loadings for the super level
        weights_block : [array_like]
            list of weights arrays of every block with all the extracted latent variables 
        weights_super : array_like
            array of all latent variables extracted weights for the super level
        c_loadings : array_like
            array of c_loadings for all extracted latent variables
        q2 : [float]
            list of r² coefficients extracted during cross validation
        feature_importances : [float]
            list of values that represent how important is each variable in the same order 
                of the X columns on the first matrix

        Methods
        -------
        fit(X, blocks_sep, Y, latent_variables, deflation, int_call)
            Applies the NIPALS like method to find the best parameters that adjust the model 
                to the data
        transform(X, latent_variables)
            Transforms the original data to the latent variable space in the super level 
        transform_inv(super_scores, latent_variables)
        transform_b(X, block, latent_variables)
            Transforms the original data to the latent variable space in the block level for
                all blocks
        predict(X, latent_variables)
            Predicts Y values 
        score(X, Y, latent_variables)
            calculates the r² value for Y
        Hotellings_T2(X)
            Calculates Hotellings_T2 values for the X data in the super level
        T2_limit(alpha)
            Returns the Hotelling's T² limit estimated with alpha confidence level
        SPEs_X(X)
            Calculates squared prediction errors on the X side for the super level
        SPEs_X_limit(X)
            Calculates squared prediction errors for the block level for all blocks
        SPEs_Y(X, Y)
            Calculates squared prediction errors for the predictions
        SPEs_Y_limit(X, Y)
            Calculates squared prediction errors for the predictions
        contributions_scores_ind(X)
            calculates the contributions of each variable to the scores on the super level
        contributions_SPE(X)
            calculates the contributions of each variable to the SPE on the X side for
                the super level

    """
    def __init__(self, cv_splits_number = 7, tol = 1e-8, loop_limit = 1000, missing_values_method = 'TSM'):
        
        # Parameters
        
        self.cv_splits_number = cv_splits_number # number of splits for latent variable cross-validation
        self.tol = tol # criteria for convergence
        self.loop_limit = loop_limit # maximum number of loops before convergence is decided to be not attainable
        self.q2y = [] # list of cross validation scores
        self.missing_values_method = missing_values_method
        
        # Attributes

        self.latent_variables = None # number of principal components to be extracted
        self.block_divs = None
        self.block_p_loadings = None
        self.superlevel_p_loadings = None
        self.block_weights = None
        self.superlevel_weights = None
        self.x_weights_star = None
        self.x_weights = None
        self.c_loadings = None 
        self.feature_importances_ = None
        self.coefficients = None
        self.omega = None # score covariance matrix for missing value estimation
        self._int_PLS = None
        self.training_scores = None
        self.training_sl_scores = None

    def fit(self, X, block_divs, Y, latent_variables = None, deflation = 'both', int_call = False):
        """
        Adjusts the model parameters to best fit the Y using the algorithm defined in 
            Westerhuis et al's [1]

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        block_divs : [int]
            list with the index of every block final position. ex. [2,4] means two blocks with 
                columns 0 and 1 on the first block and columns 2 and 3 on the second block.
        Y : array_like
            Matrix with all the data to be predicted in one only object
        latent_variables : int, optional
            Number of latent variables deemed relevant from each block. If left unspecified
                a cross validation routine will define the number during fitting
        deflation : str, optional
            string defining method of deflation, only Y or both X and Y 
        int_call : Boolean, optional
            Flag to define if it is an internal call on the cross validation routine and decide
                if it is necessary to calculate the VIP values

        References 
        ----------
        [1] J. A. Westerhuis, T. Kourti, and J. F. MacGregor, “Analysis of multiblock and hierarchical 
        PCA and PLS models,” Journal of Chemometrics, vol. 12, no. 5, pp. 301–321, 1998, 
        doi: 10.1002/(SICI)1099-128X(199809/10)12:5<301::AID-CEM515>3.0.CO;2-S.


        """
        if isinstance(X, DataFrame):# If X data is a pandas Dataframe
            X = array(X.to_numpy(), ndmin = 2)

        self.block_divs = block_divs
        block_coord_pairs = (*zip([0] + block_divs[:-1], block_divs),)
          
        if isinstance( Y, DataFrame ) or isinstance( Y, Series ): # If Y data is a pandas Dataframe or Series 
            Y = array( Y.to_numpy(), ndmin = 2 ).T
        
        #Using a full PLS as basis to calculate the MB-PLS
        int_PLS = PLS(cv_splits_number = self.cv_splits_number, tol = self.tol, loop_limit = self.loop_limit, missing_values_method = self.missing_values_method)
        int_PLS.fit(X, Y, latent_variables = latent_variables, deflation = deflation)

        

        block_coord_pairs = (*zip([0] + block_divs[:-1], block_divs),)

        superlevel_T = zeros((X.shape[0], len(block_coord_pairs) * int_PLS.latent_variables))
        superlevel_weights = zeros((int_PLS.latent_variables , len(block_coord_pairs) * int_PLS.latent_variables))
        block_weights = zeros((int_PLS.latent_variables * len(block_coord_pairs), X.shape[1]))
        block_p_loadings = zeros(block_weights.shape)

        for block, (start, end) in enumerate(block_coord_pairs):
            test_missing_data = isnan(sum(X[:, start:end]))
            if test_missing_data:
                b_weights = zeros((int_PLS.latent_variables, end - start))
                for i in range(int_PLS.latent_variables):
                    b_weights[i, start:end] = nansum(X[:, start:end] * array(int_PLS.training_scores[:, i], ndmin = 2).T, axis = 0) / nansum(((~isnan(X[:, start:end]).T * int_PLS.training_scores[:, i]) ** 2), axis = 1)
            else:
                b_weights = array(X[:, start:end].T @ int_PLS.training_scores / diagonal(int_PLS.training_scores.T @ int_PLS.training_scores), ndmin = 2).T

            block_weights[(block * int_PLS.latent_variables):((block + 1) * int_PLS.latent_variables), start:end] = b_weights
            
            block_scores = zeros((X.shape[0], int_PLS.latent_variables))
            
            
            if test_missing_data:
                for i in range(int_PLS.latent_variables):
                    block_scores[:, i] = nansum(X[:, start:end] * b_weights[i, :], axis = 1) / nansum(((~isnan(X[:, start:end]) * b_weights[i, :]) ** 2), axis = 1)
                    b_p_loadings = array(sum(nan_to_num(X) * block_scores[:, i], axis = 0) / sum((~isnan(X) * block_scores[:, i]) ** 2, axis = 0), ndmin = 2)
                    
            else:
                block_scores = (X[:, start:end] @ b_weights.T)
                b_p_loadings = zeros((int_PLS.latent_variables, end-start))
                for i in range(int_PLS.latent_variables):                  
                    b_p_loadings[i:i+1, :] = array(X[:, start:end].T @ block_scores[:, i] / (block_scores[:, i].T @ block_scores[:, i]), ndmin = 2)
            block_p_loadings[(block * int_PLS.latent_variables):((block + 1) * int_PLS.latent_variables), start:end] = b_p_loadings
            superlevel_T[:, [block + num * len(block_coord_pairs) for num, _ in enumerate(block_coord_pairs)]] = block_scores
        
        for i in range(int_PLS.latent_variables):
            numerator = (superlevel_T[:, i * len(block_coord_pairs) : (i + 1) * len(block_coord_pairs)].T @ int_PLS.training_scores[:, i])
            denominator = (int_PLS.training_scores[:, i].T @ int_PLS.training_scores[:, i])
            superlevel_weights[i , i * len(block_coord_pairs) : (i + 1) * len(block_coord_pairs)] = numerator / denominator #Error here
        
        
        """-------------------------------- p loadings calculation section ----------------------------"""                        
        

        #----------------Attribute Assignment---------------

        

        self.latent_variables = int_PLS.latent_variables
        self.block_divs = block_divs
        

        self.block_p_loadings = block_p_loadings
        self.block_weights = block_weights
        self.superlevel_weights = superlevel_weights
        self.training_sl_scores = superlevel_T

        self.x_weights_star = int_PLS.weights_star
        self.x_weights = int_PLS.weights
        self.c_loadings = int_PLS.c_loadings
        self.feature_importances_ = int_PLS.feature_importances_
        self.omega = int_PLS.omega
        self._int_PLS = int_PLS
        self.training_scores = self._int_PLS.training_scores
        
    def transform(self, X, latent_variables = None):

        """

        Transforms the X sample to the principal component space.

        Parameters
        ----------
        X : array_like
            Samples to transform
        latent_variables : int, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        result : array_like 
            returns X samples' superlevel scores.  

        """

        return self._int_PLS.transform(X, latent_variables)

    def transform_inv(self, super_scores, latent_variables = None):

        """
        Transforms the super_scores matrix to the original X.

        Parameters
        ----------
        super_scores : array_like
            Matrix with all the scores to be used to rebuild X
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        result : array_like 
            matrix of rebuilt X from scores
        """

        return self._int_PLS.transform_inv(super_scores, latent_variables = None)

    def predict(self, X, latent_variables = None):

        """
        
        Predicts Y for X samples.

        Parameters
        ----------

        X : array_like 
            Samples to use for prediction
        latent_variables : int, Optional
            Number of desired latent variables to be used. If 
        none are supplied, all variables will be used.

        Returns
        -------
        result : array_like
            Returns Y predictions

        """

        return self._int_PLS.predict(X, latent_variables = latent_variables)

    def score(self, X, Y, latent_variables = None):

        """
        Return the coefficient of determination R^2 of the prediction.

        R² is defined as 1 - Variance(Error) / Variance(Y) with Error = Y - predictions(X)

        Parameters
        ----------
        X : array_like
            Matrix with all the X to be used
        Y : array_like
            Matrix with all the Y ground truth values
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        score : float 
            returns calculated r².
        """

        return self._int_PLS.score(X, Y, latent_variables = latent_variables)

    def Hotellings_T2(self, X, latent_variables = None):
        
        """
        Calculates the Hotelling's T² for the X samples.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        T2s : array_like 
            returns all calculated T²s for the X samples
        """

        return self._int_PLS.Hotellings_T2(X, latent_variables = latent_variables)

    def T2_limit(self, alpha, latent_variables = None):
        
        """
        Calculates the Hotelling's T² limit based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        t2_limit : array_like 
            returns the limit T² for the alpha based on the training dataset
        """
        return self._int_PLS.T2_limit(alpha, latent_variables = latent_variables)
    
    def SPEs_X(self, X, latent_variables = None):
        
        """
        Calculates the Squared prediction error for for the X matrix rebuild.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE : array_like 
            returns all calculated SPEs for the X samples
        """
        return self._int_PLS.SPEs_X(X, latent_variables = latent_variables)

    def SPE_X_limit(self, alpha, latent_variables = None):
        """
        Calculates the SPE limit for the X rebuild based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_limit : array_like 
            returns the limit SPE for the alpha based on the training dataset
        """
        return self._int_PLS.SPE_X_limit(alpha, latent_variables = latent_variables)

    def SPEs_Y(self, X, latent_variables = None):
        """
        Calculates the Squared prediction error for the Y values.

        Parameters
        ----------
        X : array_like
            Samples X Matrix
        Y : array_like
            ground truth Y Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE : array_like 
            returns all calculated SPEs for the X samples
        """
        return self._int_PLS.SPEs_Y(X, latent_variables = latent_variables)
    
    def SPE_Y_limit(self, alpha, latent_variables = None):
        """
        Calculates the SPE limit for the X rebuild based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_limit : array_like 
            returns the limit SPE for the alpha based on the training dataset
        """
        
        return self._int_PLS.SPEs_Y_limit(alpha, latent_variables = latent_variables)

    def RMSEE (self, X, Y, latent_variables = None):

        """
        Returns the root mean squared prediction error.

        Parameters
        ----------
        X : array_like
            Matrix with all the X to be used
        Y : array_like
            Matrix with all the Y ground truth values
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        RMSEE : float 
            returns calculated RMSEE.
        """
        return self._int_PLS.RMSEE(X, Y, latent_variables = latent_variables)

    def contributions_scores_ind(self, X, latent_variables = None):
        """
        calculates the sample individual contributions to the scores.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        contributions : array_like 
            matrix of scores contributions for every X sample
        """
        
        return self.int_PLS.contributions_scores_ind(X, latent_variables = latent_variables)

    def contributions_SPE_X(self, X, latent_variables = None):
        """
        calculates the individual sample individual contributions to the squared prediction error 
            of the X variables matrix reconstruction.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_contributions : array_like 
            matrix of SPE contributions for every X sample
        """
        return self.int_PLS.contributions_SPE_X(X, latent_variables = latent_variables)

    def transform_b(self, X, block, latent_variables = None): # requires testing & development

        """
        
        Transform the X sample of a specific block to the block space.
        Does not handle missing data.

        Parameters 
        ----------
        X : array_like
            Samples to transform
        latent_variables : int, Optional
            Number of desired latent variables to be used. If
        none are supplied, all variables will be used.

        Returns
        -------
        result : array_like
            returns X samples' block level scores

        """

        if isinstance(X, DataFrame): X = X.to_numpy()      
        if latent_variables is None: latent_variables = self.latent_variables

        if block == 0:
            divs = [0 , self.block_divs[block]]
        else:
            divs = [self.block_divs[block - 1] , self.block_divs[block]]
        result = X @ self.block_weights[(block * latent_variables):(block * latent_variables) + latent_variables, divs[0]:divs[1]].T
        
        return result

    def transform_b_inv(self, scores, block, latent_variables = None): # requires testing & development

        """
        
        Transform the X sample of a specific block to the block space.
        Does not handle missing data.

        Parameters 
        ----------
        scores : array_like
            Sample scores to transform back to block domain
        latent_variables : int, Optional
            Number of desired latent variables to be used. If
        none are supplied, all variables will be used.

        Returns
        -------
        result : array_like
            returns X samples
        """

        if isinstance(scores, DataFrame): scores = scores.to_numpy()      
        if latent_variables is None: latent_variables = self.latent_variables

        if block == 0:
            divs = [0 , self.block_divs[block]]
        else:
            divs = [self.block_divs[block - 1] , self.block_divs[block]]
        result = scores @ self.block_weights[(block * latent_variables):(block * latent_variables) + latent_variables, divs[0]:divs[1]]
        
        return result

    def score_b(self, X, block, latent_variables = None): #requires testing & development

        """
        
        Transform the X sample of a specific block to the block space.
        Does not handle missing data.

        Parameters 
        ----------
        X : array_like
            Samples to transform
        latent_variables : int, Optional
            Number of desired latent variables to be used. If
        none are supplied, all variables will be used.

        Returns
        -------
        result : array_like
            returns X samples' block level scores

        """

        error = X - self.transform_b_inv(self.transform_b(X, block, latent_variables = latent_variables), block, latent_variables = latent_variables)
        
        result = 1 - nanvar(error) / nanvar(X)

        return result



import pandas as pd
from numpy import array, max, sqrt
import pytest
import sys
import os
sys.path.append(".")
from trendfitter.models import SMB_PLS

TESTDATA1_FILENAME = os.path.join(os.path.dirname(__file__), 'smb_pls_dataset.csv')

class TestSMBPLS(object):
    """
    def test_fit_missingdataset(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        assert test_model.latent_variables == [1, 1], 'Number of principal components is not as expected'

        block_p_loadings = array([[ 0.72187055,  0.53149537,  0.38586444,  0.18720989, -0.14682118, -0.01077914,
                                   -0.19177375,  0.73084692,  0.57978842, -0.30484587],
                                  [ 0.        ,  0.        ,  0.        ,  0.        ,  0.         , 0.        ,
                                    0.46434217,  0.8654855 ,  0.21667575, -0.01441398]])

        assert test_model.block_p_loadings == pytest.approx(block_p_loadings), 'block p loadings are not as expected'
        
        superlevel_p_loadings = array([[ 0.71366597,  0.52545454,  0.38147881,  0.18508211, -0.14515245, -0.01065663,
                                        -0.02883137,  0.10987593,  0.08716571, -0.0458307 ],
                                       [ 0.        ,  0.        ,  0.        ,  0.         , 0.         , 0.        ,
                                         0.46434217,  0.8654855 ,  0.21667575, -0.01441398]])

        assert test_model.superlevel_p_loadings == pytest.approx(superlevel_p_loadings), 'superlevel p loadings are not as expected'
        
        x_weights = array([[ 0.29286555,  0.2441086 ,  0.11274504,  0.05083408, -0.01934472, -0.01542236,
                             0.36462124,  0.82820732,  0.13915657, -0.02219426],
                           [ 0.        ,  0.        ,  0.        ,  0.       ,   0.        , 0.         ,
                             0.43009356,  0.89499372,  0.11828009, -0.0039479 ]])

        assert test_model.x_weights == pytest.approx(x_weights), 'x_weights are not as expected'

        block_weights = array([[ 0.7292824 ,  0.60786974,  0.28075331,  0.12658505, -0.04817148, -0.03840416,
                                -0.19177375,  0.73084692,  0.57978842, -0.30484587],
                               [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        , 0.         ,
                                 0.43009356,  0.89499372,  0.11828009, -0.0039479 ]])

        assert test_model.block_weights == pytest.approx(block_weights), 'block_weights are not as expected'

        x_weights_star = array([[ 0.71448381,  0.5260567 ,  0.38191598,  0.18529421, -0.14531879, -0.01066884,
                                 -0.07538501,  0.02329228,  0.06555772, -0.04443914],
                                [-0.07149937, -0.05264321, -0.03821885, -0.01854264,  0.01454225,  0.001067645,
                                  0.46645486,  0.85303144,  0.20758095, -0.0097983 ]])

        assert test_model.x_weights_star == pytest.approx(x_weights_star), 'x_weights_star are not as expected'

        superlevel_weights = array([[0.98863427, 0.15034055],
                                    [0.        , 1.        ]])

        assert test_model.superlevel_weights == pytest.approx(superlevel_weights), 'superlevel weights is not as expected'

        c_loadings = array([[1.11716365],
                            [1.65955812]])

        assert test_model.c_loadings == pytest.approx(c_loadings), 'c_loadings is not as expected'

        expected_q2 = array([0.270841981393722, 0.9249127776764678])

        assert test_model.q2y == pytest.approx(expected_q2), 'Q2 results are not the expected'
    
    def test_fit_fulldataset(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        assert test_model.latent_variables == [1, 2], 'Number of principal components is not as expected'

        block_p_loadings = array([[ 0.72434131,  0.54728295,  0.39237194,  0.16893039, -0.16190754, -0.02024584,
                                   -0.04310647,  0.76757319,  0.59608966, -0.23162544],
                                  [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
                                    0.46522025,  0.86612955,  0.21192985, -0.01607839],
                                  [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
                                   -0.32421554,  0.2704348 , -0.89330915,  0.1636237 ]])

        assert test_model.block_p_loadings == pytest.approx(block_p_loadings), 'block p loadings are not as expected'
        
        superlevel_p_loadings = array([[ 0.71675602,  0.54155181,  0.38826303,  0.16716135, -0.16021205, -0.02003382,
                                        -0.00622203,  0.11079223,  0.08604014, -0.03343303],
                                       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
                                         0.46522025,  0.86612955,  0.21192985, -0.01607839],
                                       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
                                        -0.32421554,  0.2704348 , -0.89330915,  0.1636237 ]])

        assert test_model.superlevel_p_loadings == pytest.approx(superlevel_p_loadings), 'superlevel p loadings are not as expected'
        
        x_weights = array([[ 0.28902875,  0.24309712,  0.11359796,  0.04388081, -0.0178012,  -0.01825811,
                             0.37337076,  0.82711786,  0.13439652, -0.01714089],
                           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.       ,   0.        ,
                             0.4301038 ,  0.89551408,  0.11422467, -0.00424076],
                           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.       ,   0.        ,
                            -0.32350558,  0.27070105, -0.90009589,  0.10905257]])

        assert test_model.x_weights == pytest.approx(x_weights), 'x_weights are not as expected'

        block_weights = array([[ 0.72686835,  0.61135648,  0.28568356,  0.11035432, -0.04476762, -0.0459167,
                                -0.04310647,  0.76757319,  0.59608966, -0.23162544],
                               [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,  0.       ,
                                 0.4301038 ,  0.89551408,  0.11422467, -0.00424076],
                               [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,  0.       ,
                                -0.32350558,  0.27070105, -0.90009589,  0.10905257]])

        assert test_model.block_weights == pytest.approx(block_weights), 'block_weights are not as expected'

        x_weights_star = array([[ 7.03552842e-01,  5.31576028e-01,  3.81110933e-01,  1.64082113e-01,
                                 -1.57260822e-01, -1.96647870e-02, -6.69998387e-02,  2.85750481e-02,
                                  2.84020306e-02, -2.49247641e-02],
                                [-7.48473945e-02, -5.65516593e-02, -4.05444462e-02, -1.74558582e-02,
                                  1.67301759e-02,  2.09203628e-03,  4.37245350e-01,  8.92324884e-01,
                                  1.12228140e-01,  4.28092118e-03],
                                [ 2.72179697e-02,  2.05647954e-02,  1.47438334e-02,  6.34775631e-03,
                                 -6.08386470e-03, -7.60761019e-04, -2.79281481e-01,  3.67623305e-01,
                                 -8.77043981e-01,  1.62342367e-01]])

        assert test_model.x_weights_star == pytest.approx(x_weights_star), 'x_weights_star are not as expected'

        superlevel_weights = array([[0.98952802, 0.14434094],
                                    [0.        , 1.        ],
                                    [0.        , 1.        ]])

        assert test_model.superlevel_weights == pytest.approx(superlevel_weights), 'superlevel weights is not as expected'

        c_loadings = array([[1.13247757],
                            [1.66200662],
                            [0.19718822]])

        assert test_model.c_loadings == pytest.approx(c_loadings), 'c_loadings is not as expected'

        expected_q2 = array([0.28885562322442215, 0.9301799477729965, 0.9413869033563421])

        assert test_model.q2y == pytest.approx(expected_q2), 'Q2 results are not the expected'
    
    def test_transform_missingdataset(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        scores = array([[ 0.3180861 ,  0.40284835],
                        [-0.28374995, -0.71410348],
                        [-0.27569826, -1.01515561],
                        [-0.27132141, -0.32671546],
                        [-0.37764758,  0.25349725]])

        assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function malfunctions for datasets with missing values'

    def test_transform_fulldataset(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        scores = array([[ 0.29239167,  0.38514696, -0.1151795 ],
                        [-0.2536207 , -0.66350758,  0.45991813],
                        [-0.28589818, -1.02426442, -0.14611273],
                        [-0.27384959, -0.36088289, -0.29182376],
                        [-0.37422809,  0.19158321, -0.57872799]])

        assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_transform_b_fulldataset(self):#To Do
        
        a=1 

    def test_predict(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        predictions = array([[ 0.94853176],
                             [-1.29928331],
                             [-2.05491923],
                             [-0.96746247],
                             [-0.2195107 ]])

        assert test_model.predict(X_test_data.iloc[:5]) == pytest.approx(predictions), 'The predict function for perfect datasets malfunctions'
    
    def test_score(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        assert test_model.score(X_test_data.iloc[:5], Y_test_data.iloc[:5]) == pytest.approx(0.9706446197457211), 'The score function for perfect datasets malfunctions'

    def test_Hotellings_T2(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        T2s = array([[1.61028906],
                     [4.66663225],
                     [7.50870743],
                     [1.89242678],
                     [3.33450584]])

        assert test_model.Hotellings_T2(X_test_data.iloc[:5]) == pytest.approx(T2s), 'The Hotellings_T2 function for perfect datasets malfunctions'

    def test_Hotellings_T2_blocks(self): #To Do

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)  

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_SPEs_X(self): # To Do
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_SPEs_Y(self): # To Do

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_contributions_scores(self): # To Do

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'
    
    def test_contributions_scores_block(self): # To Do
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_contributions_SPE_X(self): # To Do
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_contributions_SPE_X_block(self): # To Do
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'
    """
    

    




from numpy import min, sum, mean, std, nanvar, insert, array, where, isnan, nan_to_num, nansum, nanmean, identity, zeros, unique, var, append, nansum, any
from numpy.linalg import norm
from pandas import DataFrame, Series
from sklearn.model_selection import KFold
from scipy.stats import f, chi2
from trendfitter.auxiliary.tf_aux import scores_with_missing_values



class PLS:

    """
    A sklearn-like class for the Projection to Latent Structures.

        Parameters
        ----------
        cv_splits_number : int, optional
            number of splits used for cross validation in case latent_variables is None
        tol : float, optional
            value used to decide if model has converged
        loop_limit : int, optional
            maximum number of loops for the extraction of one latent variable
        missing_values_method : str, optional
            string to define the method that the model will deal with missing values

        Attributes
        ----------
        latent_variables : int
            number of latent variables one wants to extract.
        p_loadings_block : [array_like]
            list of p loadings arrays of every block with all the extracted latent variables 
        weights : array_like
            array of all latent variables extracted weights 
        weights_star : array_like
            array of all latent variables extracted weights for the non-deflated X matrix
        c_loadings : array_like
            array of c_loadings for all extracted latent variables
        q2 : [float]
            list of r² coefficients extracted during cross validation
        feature_importances_ : [float]
            list of values that represent how important is each variable in the same order 
                of the X columns on the first matrix
        omega : array_like
            If missing_values_method requires a scores covariance matrix ('TSR', 'CMR', 'PMP'), 
                it will be stored here.
        training_scores : array_like
            Scores extracted during training of the model.

        Methods
        -------
        fit(X, Y, latent_variables = None, deflation = 'both', random_state = None, int_call = False)
            Applies the NIPALS like method to find the best parameters that adjust the model 
                to the data
        transform(X, latent_variables = None)
            Transforms the original data to the latent variable space in the super level 
        transform_inv(scores, latent_variables = None)
            Returns the scores to the original data space
        predict(X, latent_variables = None)
            Predicts Y values 
        score(X, Y, latent_variables = None)
            calculates the r² value for Y
        Hotellings_T2(X, latent_variables = None)
            Calculates Hotellings_T2 values for the X data in the super level
        T2_limit(alpha, latent_variables = None)
            Returns the Hotelling's T² limit estimated with alpha confidence level
        SPEs_X(X, latent_variables = None)
            Calculates squared prediction errors on the X side
        SPE_X_limit(alpha, latent_variables = None)
            Returns the squared prediction error limit with alpha confidence level 
        SPEs_Y(X, Y, latent_variables = None)
            Calculates squared prediction errors for the predictions
        SPE_Y_limit(alpha, latent_variables = None)
            Returns the squared prediction error limit with alpha confidence level
        RMSEE(X, Y, latent_variables = None)
            returns the Root mean square estimation error
        contributions_scores_ind(X, latent_variables = None)
            calculates the contributions of each variable to the scores on the super level
        contributions_SPE_X(X, latent_variables = None)
            calculates the contributions of each variable to the SPE on the X side for 

    """
    def __init__(self, cv_splits_number = 7, tol = 1e-8, loop_limit = 1000, missing_values_method = 'TSM'):
        
        self.p_loadings = None
        self.weights = None
        self.weights_star = None
        self.c_loadings = None
        self.latent_variables = None # number of principal components to be extracted
        self.cv_splits_number = cv_splits_number # number of splits for latent variable cross-validation
        self.tol = tol # criteria for convergence
        self.loop_limit = loop_limit # maximum number of loops before convergence is decided to be not attainable
        self.q2y = [] # list of cross validation scores
        self.deflation = None
        self.coefficients = None
        self.training_scores = None
        self.omega = None
        self.missing_values_method = missing_values_method
        self.feature_importances_ = None #for scikit learn use with feature selection methods
        self._x_chi2_params = []
        self._y_chi2_params = []
               
    def fit(self, X, Y, latent_variables = None, deflation = 'both', random_state = None, int_call = False):

        """
        Adjusts the model parameters to best fit the Y using the algorithm defined in 
            Wold's [1]

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        Y : array_like
            Matrix with all the data to be predicted in one only object
        latent_variables : int, optional
            number of latent variables deemed relevant. If left unspecified
                a cross validation routine will define the number during fitting.
        deflation : str, optional
            string defining method of deflation, only Y or both X and Y 
        random_state : int, optional
            value used as a seed in the random number generator for cross validation

        References 
        ----------
        [1] S. Wold, M. Sjöström, and L. Eriksson, “PLS-regression: a basic tool of chemometrics,” 
        Chemom. Intell. Lab. Syst., vol. 58, no. 2, pp. 109–130, Oct. 2001, doi: 10.1016/S0169-7439(01)00155-1.


        """


        """----------------------- Dealing with data possibly coming in the form of a pandas dataframe ------------------------"""
        
        self.latent_variables = latent_variables
        self.deflation = deflation
        
        if isinstance(X, DataFrame): # in case the function receives a dataframe as X data
            X = array(X.to_numpy(), ndmin = 2)
        Orig_X = X
            
        if isinstance(Y, DataFrame): Y = array( Y.to_numpy(), ndmin = 2)
        elif isinstance(Y, Series): Y = array( Y.to_numpy(), ndmin = 2).T # in case the function receives a dataframe as Y data
        else: Y = array(Y, ndmin = 2)
        Orig_Y = Y
    
        """------------------- Checking if the data is in fact problematic -------------------------"""

        if isnan(sum(X)) or isnan(sum(Y)):
            dataset_complete = False
        else: dataset_complete = True
        
        """------------------- Handling the case where the amount of latent variables is not defined -------------------"""
        if self.latent_variables != None:
            latent_variables = self.latent_variables
        else:
            latent_variables = min(X.shape) #maximum amount of extractable latent variables
            kf = KFold(n_splits = self.cv_splits_number, shuffle = True , random_state = random_state)

        """------------------- model Calculation -------------"""
        G = identity(X.shape[1])
        q2_final = []
        for latent_variable in range(1, latent_variables + 1):
            
            loop_count = 0
            conv = 1
            
            y_scores1 = array(nan_to_num( Y[:,0] ), ndmin = 2).T #handling possible NaNs that come from faulty datasets - Step 1

            while conv > self.tol and loop_count < self.loop_limit:#NIPALS internal Loop 
                
                if dataset_complete: # if the dataset is without problematic data, matrix implementation is possible and faster
                    
                    weights = array(X.T @ y_scores1 / (y_scores1.T @ y_scores1), ndmin = 2).T #calculating w vector by regressing all u values into X transpose
                    weights = weights / norm(weights) #normalizing w vector
                    x_scores = X @ weights.T / (weights @ weights.T) #regressing all w values into X matrix and producing a t matrix
                    c_loadings = array(Y.T @ x_scores / (x_scores.T @ x_scores), ndmin = 2).T #regressing all t values into Y and producing the c matrix
                    y_scores = Y @ c_loadings.T / (c_loadings @ c_loadings.T) #calculating new u values by regressing cs into Ys
                    
                    conv = norm(y_scores1 - y_scores)
                    y_scores1 = y_scores
                    
                else: # if it isn't without problematic data, then one needs to deal with it by looping for every regression
                                        
                    """ This segment of the code is a lot slower. 
                    - An opportunity to make it faster is to separate the lines and columns
                    with missing data from the ones complete and iterate only on the ones with missing data. The ones with complete lines can 
                    be used for one single matrix operation. 
                    - A second opportunity is to group the missing data columns and lines based on which variables are missing to then execute a matrix 
                    operation for each group."""
                       
                    weights = array(nansum((X * y_scores1), axis = 0) / nansum(((~isnan(X) * y_scores1) ** 2).T, axis = 1), ndmin = 2) #calculating w vector by regressing all u values into X transpose
                    weights = weights / norm(weights) #normalizing w vector
                    x_scores = array(nansum(X * weights, axis = 1) / nansum(((~isnan(X) * weights) ** 2), axis = 1), ndmin = 2).T #regressing all w values into X matrix and producing a t matrix
                    c_loadings = array(nansum(Y * x_scores, axis = 0) / nansum(((~isnan(Y) * x_scores) ** 2).T, axis = 1), ndmin = 2) #regressing all t values into Y and producing the c matrix
                    y_scores =  array(nansum(Y * c_loadings, axis = 1) / nansum(((~isnan(Y) * c_loadings) ** 2), axis = 1), ndmin = 2).T #calculating new u values by regressing cs into Ys

                    conv = norm(y_scores1 - y_scores)
                    y_scores1 = y_scores
                    
               
                loop_count += 1
                if loop_count == self.loop_limit: 
                    print(f'Model could not converge with {self.loop_limit} loops')
                    return
                
            
            
            """-------------------------------- Cross validation section -----------------------------------"""
            if self.latent_variables is None: # if the quantity of principal components is undefined, a cross-validation routine decides when to stop
                
                testq2 = []
                

                q2_model = PLS(tol = self.tol, loop_limit = self.loop_limit, missing_values_method = 'TSM') # an internal model is initialized
                for train_index, test_index in kf.split(X):
                    q2_model.fit(Orig_X[train_index, :], Orig_Y[train_index, :], latent_variables = latent_variable, deflation = self.deflation, int_call = True) # this model is trained with only a partition of the total dataset
                    testq2.append(q2_model.score(Orig_X[test_index, :], Orig_Y[test_index, :])) # its performance is registered in a list
                q2_final.append(mean(testq2))
                """ ------------------ coefficients and VIP calculations ----------------- """
             
                if latent_variable > 1: 
                    
                    if (q2_final[-1] < q2_final[-2] or  # mean cross-validation performance of the model has decreased with the addition of another latent variable
                        q2_final[-1] - q2_final[-2] < 0.01 or  # mean cross-validation performance of the model has increased less than 1% with the addition of another latent variable
                        latent_variable > min(X.shape) / 2): # the amount of latent variables added is more than half the variables on X
                        self.q2y = q2_final[:-1]
                        self.latent_variables = latent_variable - 1 
                        if self.missing_values_method == 'TSM': break  #In case of TSM use there is no need of more components for missing value estimation
                        
            """-------------------------------- p loadings calculation section ----------------------------"""                        
            if dataset_complete: # if the dataset is without problematic data, matrix implementation is possible and faster
                p_loadings = array(X.T @ x_scores / (x_scores.T @ x_scores), ndmin = 2).T
                
            else: # if it isn't without problematic data, then one needs to deal with it
                p_loadings = array(sum(nan_to_num(X) * x_scores, axis = 0) / sum((~isnan(X) * x_scores) ** 2, axis = 0), ndmin = 2)

            """-------------------------------- deflation section -----------------------------------------"""
            if self.deflation == 'both':                

                X = X - x_scores @ p_loadings
                Y = Y - x_scores @ c_loadings

                SPE_X = sum(X[~(any(isnan(X), axis = 1))] ** 2, axis = 1)
                SPE_Y = sum(Y[~(any(isnan(Y), axis = 1))] ** 2, axis = 1)
                self._x_chi2_params.append((2 * mean(SPE_X) ** 2) / var(SPE_X)) #for future SPE analysis
                self._y_chi2_params.append((2 * mean(SPE_Y) ** 2) / var(SPE_Y)) #for future SPE analysis
                
            elif self.deflation == 'Y':

                SPE_Y = sum(Y[~(any(isnan(Y), axis = 1))] ** 2 , axis = 1)
                
                SPE_Y = sum(Y ** 2, axis = 1)
                self._y_chi2_params.append((2 * mean(SPE_Y) ** 2) / var(SPE_Y)) #for future SPE analysis

            else:
                raise ValueError("impossible deflation parameter")
            
            
            
            """-------------------------------- class property assignment section -------------------------"""
            if latent_variable < 2:
                self.p_loadings = p_loadings
                self.weights = weights
                self.c_loadings = c_loadings
                self.weights_star = weights
                G = G - weights.T @ p_loadings 
                self.training_scores = array(x_scores, ndmin = 2)
            else:
                weights_star = array(G @ weights.T , ndmin = 2).T
                G = G - weights_star.T @ p_loadings 
                self.weights_star = insert(self.weights_star, self.weights_star.shape[0], weights_star, axis = 0)          
                self.p_loadings = insert(self.p_loadings, self.p_loadings.shape[0], p_loadings, axis = 0)
                self.weights = insert(self.weights, self.weights.shape[0], weights, axis = 0)
                self.c_loadings = insert(self.c_loadings, self.c_loadings.shape[0], c_loadings, axis = 0)
                self.training_scores = insert(self.training_scores, self.training_scores.shape[1], array( x_scores, ndmin = 2).T, axis = 1)
            self.omega = self.training_scores.T @ self.training_scores 

            """-------------------------------- Coefficients property calculations ----------------"""
           
            #coefficients_calc = self.weights_star.dot( self.c_loadings.T )
            
            """-------------------------------- VIPs property calculations ------------------------"""
            
            if self.latent_variables != None and latent_variable > 2 * self.latent_variables: break
        #VIP calculation
        if not int_call: self.feature_importances_ = self._VIPs_calc(Orig_X,Orig_Y)

        return
        
    def transform(self, X, latent_variables = None):

        """
        Transforms the X matrix to the model-fitted space.

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        result : array_like of shape (X.shape[0], latent_variables)
            Scores for the X values transformed

        """
        
        if isinstance(X, DataFrame): X = X.to_numpy()      
        if latent_variables is None: latent_variables = self.latent_variables
        
        if isnan(sum(X)):
            
            result = zeros((X.shape[0], latent_variables))
            X_nan = isnan(X)
            variables_missing_mask = unique(X_nan, axis = 0)

            for row_mask in variables_missing_mask:
                
                rows_indexes = where((X_nan == row_mask).all(axis = 1))                
                
                if sum(row_mask) == 0: 

                    result[rows_indexes, :] = X[rows_indexes, :] @ self.weights_star[:latent_variables, :].T 
                
                else:
                    
                    result[rows_indexes, :] = scores_with_missing_values(self.omega, self.weights_star[:, ~row_mask], X[rows_indexes[0][:, None], ~row_mask], 
                                                                            LVs = latent_variables, method = self.missing_values_method)
                    
        else: result = X @ self.weights_star[:latent_variables, :].T 

        return result
    
    def transform_inv(self, scores, latent_variables = None):

        """
        Transforms the scores matrix to the original X.

        Parameters
        ----------
        scores : array_like
            Matrix with all the scores to be used to rebuild X
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        result : array_like 
            matrix of rebuilt X from scores
        """
                  
        if latent_variables is None: latent_variables = self.latent_variables
        result = scores @ self.p_loadings[:latent_variables, :] 
        
        return result
    
    def predict(self, X, latent_variables = None):
     
        """
        Predicts Y values using X array.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        preds : array_like 
            returns predictions
        """

        if isinstance(X, DataFrame): X = X.to_numpy()        
        if latent_variables is None: latent_variables = self.latent_variables        
        preds = self.transform(X, latent_variables = latent_variables) @ self.c_loadings[:latent_variables, :] 
        
        return preds
    
    def score(self, X, Y, latent_variables = None):
        
        """
        Return the coefficient of determination R^2 of the prediction.

        R² is defined as 1 - Variance(Error) / Variance(Y) with Error = Y - predictions(X)

        Parameters
        ----------
        X : array_like
            Matrix with all the X to be used
        Y : array_like
            Matrix with all the Y ground truth values
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        score : float 
            returns calculated r².
        """

        if isinstance(Y, DataFrame) or isinstance(Y, Series): Y = array(Y.to_numpy(), ndmin = 2).T     
        if latent_variables is None: latent_variables = self.latent_variables 

        Y_hat = self.predict(X, latent_variables = latent_variables)
        F = Y - Y_hat
        score = 1 - nanvar(F) / nanvar(Y)
        
        return score
    
    def Hotellings_T2(self, X, latent_variables = None):

        """
        Calculates the Hotelling's T² for the X samples.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        T2s : array_like 
            returns all calculated T²s for the X samples
        """
        
        if isinstance(X, DataFrame): X = X.to_numpy()     #dataframe, return it
   
        if latent_variables is None: latent_variables = self.latent_variables # Unless specified, the number of PCs is the one in the trained model 
        
        scores_matrix = self.transform(X, latent_variables = latent_variables)
        
        T2s = sum(((scores_matrix / std(self.training_scores[:, :latent_variables], axis = 0)) ** 2), axis = 1)
        
        return T2s
    
    def T2_limit(self, alpha, latent_variables = None):
        
        """
        Calculates the Hotelling's T² limit based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        t2_limit : array_like 
            returns the limit T² for the alpha based on the training dataset
        """

        if latent_variables is None: latent_variables = self.latent_variables # Unless specified, the number of PCs is the one in the trained model 

        F_value = f.isf(1 - alpha , latent_variables, self.training_scores.shape[0])
        t2_limit = ((latent_variables * (self.training_scores.shape[0] ** 2 - 1)) / 
                    (self.training_scores.shape[0] * (self.training_scores.shape[0] - latent_variables))) * \
                    F_value

        return t2_limit

    def SPEs_X(self, X, latent_variables = None):

        """
        Calculates the Squared prediction error for for the X matrix rebuild.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE : array_like 
            returns all calculated SPEs for the X samples
        """
        
        if latent_variables is None: latent_variables = self.latent_variables

        if isinstance(X, DataFrame): X = X.to_numpy()       
        
        error = X - self.transform_inv(self.transform(X, latent_variables = latent_variables), latent_variables = latent_variables)  
        SPE = nansum(error ** 2, axis = 1)
        
        return SPE

    def SPE_X_limit(self, alpha, latent_variables = None):

        """
        Calculates the SPE limit for the X rebuild based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_limit : array_like 
            returns the limit SPE for the alpha based on the training dataset
        """
        if self.deflation == 'Y':  raise ValueError("Impossible to extract X SPE limits with only Y deflation")

        if latent_variables is None: latent_variables = self.latent_variables # Unless specified, the number of PCs is the one in the trained model 
        
        chi2_val = chi2.isf(1 - alpha, latent_variables - 1)
        SPE_limit = self._x_chi2_params[latent_variables - 1] * chi2_val
        
        return SPE_limit

    def SPEs_Y(self, X, Y, latent_variables = None):

        """
        Calculates the Squared prediction error for the Y values.

        Parameters
        ----------
        X : array_like
            Samples X Matrix
        Y : array_like
            ground truth Y Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE : array_like 
            returns all calculated SPEs for the X samples
        """
        
        if latent_variables is None: latent_variables = self.latent_variables

        if isinstance(X, DataFrame): X = X.to_numpy()

        if isinstance(Y, DataFrame) : 
            Y = Y.to_numpy()
            Y = array(Y, ndmin = 2)
        elif isinstance(Y, Series):
            Y = Y.to_numpy()
            Y = array(Y, ndmin = 2).T
        
        
        error = Y - self.predict(X, latent_variables = latent_variables)
        SPE = nansum(error ** 2, axis = 1)
        
        return SPE
    
    def SPE_Y_limit(self, alpha, latent_variables = None):

        """
        Calculates the SPE limit for the X rebuild based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_limit : array_like 
            returns the limit SPE for the alpha based on the training dataset
        """

        if latent_variables is None: latent_variables = self.latent_variables # Unless specified, the number of PCs is the one in the trained model 
        
        chi2_val = chi2.isf(1 - alpha, latent_variables - 1)
        SPE_limit = self._y_chi2_params[latent_variables - 1] * chi2_val
        
        return SPE_limit
    
    def RMSEE (self, X, Y, latent_variables = None):

        """
        Returns the root mean squared prediction error.

        Parameters
        ----------
        X : array_like
            Matrix with all the X to be used
        Y : array_like
            Matrix with all the Y ground truth values
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        RMSEE : float 
            returns calculated RMSEE.
        """
        
        if latent_variables is None: latent_variables = self.latent_variables
        if isinstance(Y, DataFrame): Y = array(Y.to_numpy(), ndmin = 2)
        elif isinstance(Y, Series): Y = array(Y.to_numpy(), ndmin = 2).T # in case the function receives a dataframe as Y data
        else: Y = array(Y, ndmin = 2)

        Y_hat = self.predict(X)
        error = sum((Y - Y_hat) ** 2, axis = 0)
        RMSEE = (error / (Y.shape[0] - latent_variables - 1)) ** (1 / 2)  # il faut ajouter une manière de calculer multiples Y simultaneement
        
        return RMSEE
      
    def contributions_scores_ind(self, X, latent_variables = None): 

        """
        calculates the sample individual contributions to the scores.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        contributions : array_like 
            matrix of scores contributions for every X sample
        """

        if latent_variables is None: latent_variables = self.latent_variables
        if isinstance(X, DataFrame): X = X.to_numpy()

        scores = self.transform(X, latent_variables = latent_variables)
        scores = (scores / std(self.training_scores[:, :latent_variables], axis = 0)) ** 2  
        contributions = X * ((scores @ (self.weights_star[:latent_variables, :] ** 2)) ** (1 / 2))

        return contributions
    
    def contributions_SPE_X(self, X, latent_variables = None):

        """
        calculates the individual sample individual contributions to the squared prediction error 
            of the X variables matrix reconstruction.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_contributions : array_like 
            matrix of SPE contributions for every X sample
        """

        if latent_variables is None: latent_variables = self.latent_variables
        if isinstance(X, DataFrame): X = X.to_numpy()
        
        error = X - self.transform_inv(self.transform(X))
        
        SPE_contributions = (error ** 2) * where(error > 0, 1, -1)
               
        return SPE_contributions

    def _VIPs_calc( self, X, Y, latent_variables = None ): 

        """
        Calculates the VIP scores for all the variables for the prediction
        """
        if latent_variables is None: latent_variables = self.latent_variables
        
        SSY = array(sum((Y - nanmean(Y, axis = 0)) ** 2))
        for i in range(1, latent_variables + 1):
            pred = self.predict(X, latent_variables = i)
            res = Y - pred
            SSY = append(SSY, sum(((res - res.mean(axis = 0)) ** 2)))
            
        SSYdiff = SSY[:-1] - SSY[1:]
        VIPs = (((self.weights[:latent_variables, :].T ** 2) @ SSYdiff) * self.weights.shape[1] / (SSY[0] - SSY[-1]) ** 1 / 2)
       
        return VIPs


from numpy import min, sum, mean, std, var, insert, array, multiply, where, zeros, append, isnan, nan_to_num, nansum, nanvar, nanmean, unique, ix_, nonzero, nan
from numpy.linalg import norm
from scipy.stats import f, chi2
from pandas import DataFrame
from sklearn.model_selection import KFold
from trendfitter.auxiliary.tf_aux import scores_with_missing_values


class PCA:
    """
    A sklearn-like class for the NIPALs algorithm for PCA together with a toolset for 
    investigation.
    
    Parameters
    ---------- 
    tol : float, Optional
        Value used to decide if model has converged.
    loop_limit : int, Optional
        Maximum number of loops before forced stop. Resets every new component.
    missing_values_method : str, Optional
        Defines which method will be used to evaluate missing values in future transformations.
    keep_scores : boolean, Optional
        Decision to save scores extracted during model fitting. If not given, assumed False.

    Attributes
    ----------
    principal_components : int, optional
        Number of principal components extracted or to extract. If not given, a cross-validation
            internal routine will decide this value.
    cv_splits_number : int, optional
        Number of splits used for cross-validation. If not given, it will be 7.
    loadings : array_like
        Loading parameters that define the PCA model.
    q2 : [float]
        Average score on the test sets during the cross-validation procedure
    feature_importances_ : array_like
        An array that describes the importance of each feature used to build the model
            using the VIP value of each.
    training_scores : array_like
        Scores extracted during training of the model.
    omega : array_like
        If missing_values_method requires a scores covariance matrix ('TSR', 'CMR', 'PMP'), 
            it will be stored here.

    Methods
    -------
    fit(X, principal_components = None, cv_splits_number = 7, int_call = False)
        Runs the NIPALS algorithm to extract the principal components from X. 
    predict(X, principal_components = None)
        Uses the model to reconstruct X using the principal components.
    transform(X, principal_components = None)
        Transforms the X from its original space to the latent variable space.
    score(X, principal_components = None)
        Returns a r² representing how much variability from X is captured in the model. 
    Hotellings_T2(X, principal_components = None)
        Returns an array with the Hotelling's T² calculated for each row in X.
    T2_limit(alpha)
        Returns the Hotelling's T² limit estimated with alpha confidence level
    contributions_scores_ind(X, principal_components = None)
        Returns an array with the contributions to the scores of each X row.
    contributions_spe(X, principal_components = None)
        Returns an array with the contributions to the SPE of each X row.
    SPEs(X, principal_components = None)
        Returns the squared prediction errors of each row's X reconstruction.
    SPE_limit(alpha)
        Returns the squared prediction error limit with alpha confidence level
    """

    def __init__(self, tol = 1e-12, loop_limit = 100, missing_values_method = 'TSM'):
           
        self.principal_components = None # number of principal components to be extracted
        self.cv_splits_number = None # number of splits for latent variable cross-validation
        self.tol = tol # criteria for convergence
        self.loop_limit = loop_limit # maximum number of loops before convergence is decided to be not attainable
        self.missing_values_method = missing_values_method 

        self.loadings = None #loadings 
        self.q2 = [] # list of cross validation scores
        self.feature_importances_ = None #for scikit learn use with feature selection methods
        self.omega = None # scores covariance matrix for missing values score estimation
        self.training_scores = None
        self._training_scores_stds = None
        self._chi2_params = []

    def fit(self, X, principal_components = None, cv_splits_number = 7, int_call = False):
        """
        Extracts the model parameters using the NIPALs algorithm [1].

        Parameters
        ----------
        X : array_like
            Data used to extract the parameters and fit the model
        principal_components : array_like, Optional
            Number of desired principal components to be extracted
        cv_splits_number : int, Optional
            Number of desired splits to be used during the cross-validation routine        
        int_call : Boolean, optional
            Flag to determine if the method should calculate certain values. If not specified
            
        References 
        [1] S. Wold, K. Esbensen, and P. Geladi, “Principal component analysis,” 
            Chemometrics and Intelligent Laboratory Systems, vol. 2, no. 1–3, 
            pp. 37–52, Aug. 1987, doi: 10.1016/0169-7439(87)80084-9.

        """

        self.principal_components = principal_components # number of principal components to be extracted
        self.cv_splits_number = cv_splits_number # number of splits for latent variable cross-validation


        if isinstance(X, DataFrame): X = X.to_numpy() #ensuring data in the numpy format
        dataset_incomplete = False 
        if isnan(sum(X)): dataset_incomplete = True #checking if there is missing data on the input X

        #CV = False
        if not(self.principal_components is None) : 
            numberOfLVs = self.principal_components 
            #CV = True
        else :
            numberOfLVs = X.shape[1] #maximum amount of extractable latent variables
            kf = KFold(n_splits = self.cv_splits_number, shuffle = True, random_state = 1)

        MatrixXModel = array(zeros(X.shape), ndmin = 2) #initializing matrix space for manipulation 
        
        #----------------------------------------------------------------------------------
        #------------------------------NIPALS implementation-------------------------------
        #----------------------------------------------------------------------------------
        q2_final = []
        for latent_variable in range(1, numberOfLVs + 1) :
            scores_vec = nan_to_num(array(X[ :, 1], ndmin = 2).T) #initializing the guess by using a specific vector
            MatrixX2 = X - MatrixXModel #deflation of the X matrix
            counter = 0
            conv = 1
            while conv > self.tol and counter < self.loop_limit:
                counter += 1
                
                if dataset_incomplete:

                    loadings_vec = array(nansum(MatrixX2 * scores_vec, axis = 0) / nansum(((~isnan(X) * scores_vec) ** 2).T, axis = 1), ndmin = 2) # equivalent to loadings = scores'*data/scores'*scores
                    loadings_vec = loadings_vec / norm(loadings_vec) #normalization of loading vector
                    new_scores_vec = array(nansum(MatrixX2 * loadings_vec, axis = 1) / nansum(((~isnan(X) * loadings_vec) ** 2), axis = 1), ndmin = 2).T #scores calculation w/ missing data
                
                else:
                    
                    loadings_vec = array(scores_vec.T @ MatrixX2 / (scores_vec.T @ scores_vec), ndmin = 2) #loadings = scores'*data/scores'*scores        
                    loadings_vec = loadings_vec / norm(loadings_vec) #normalization of loading vector
                    new_scores_vec = MatrixX2 @ loadings_vec.T #scores calculation
                
                conv = sum((scores_vec - new_scores_vec) ** 2)  #scores comparation in between loops to assess convergency
                scores_vec = new_scores_vec # old t becomes new t

            #After convergency, if the principal components desired quantity is undefined
            #then we check if the Component is in fact significant and keep on until all 
            #components are there            
            if self.principal_components is None :
                
                testq2 = []
                
                for train_index, test_index in kf.split(X):

                    q2_model = PCA(missing_values_method = self.missing_values_method)
                    q2_model.fit(X[train_index], principal_components = latent_variable, int_call = True)
                    testq2.append(q2_model.score(X[test_index]))

                q2_final.append(mean(testq2))
                
                if latent_variable > 1:
                    
                    if (q2_final[-1] < q2_final[-2] or \
                        q2_final[-1] - q2_final[-2] < 0.01 or \
                        latent_variable > min(X.shape) / 2):
                        self.q2 = q2_final[:-1]
                        self.principal_components = latent_variable - 1
                        if self.missing_values_method == 'TSM': break 
                        
            #if significant, then we add them to the loadings and score matrixes that will be returned as method result

            #prediction of this model
            

            if latent_variable < 2:

                self.loadings = loadings_vec
                self.training_scores = scores_vec
                
            else:

                self.loadings = insert(self.loadings, self.loadings.shape[0], loadings_vec, axis = 0)
                self.training_scores = insert(self.training_scores, self.training_scores.shape[1], scores_vec.T, axis = 1)

            if not (self.principal_components) is None and latent_variable > 2 * self.principal_components: break # Ensuring to extract at least double the useful components for missing values estimation:

            MatrixXModel = self.training_scores @ self.loadings 
            error = nan_to_num(X) - MatrixXModel
            SPE = sum(error ** 2 , axis = 1)
            self._chi2_params.append([mean(SPE), var(SPE)]) #for future SPE analysis

        if not int_call: 
            self.feature_importances_ = self._VIPs_calc(X,  principal_components = self.principal_components)
            self.omega = self.training_scores.T @ self.training_scores # calculation of the covariance matrix
            self._training_scores_stds = std(self.training_scores, axis = 0)
        pass
        
    def predict(self, X, principal_components = None):

        """
        Transforms the X sample to the principal component space and back to evaluate what is
        the model "prediction" of the original sample values.

        Parameters
        ----------
        X : array_like
            Data used to extract the parameters and fit the model
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        preds : array_like 
            returns "predicted" X values.        
        """

        if isinstance(X, DataFrame): X = X.to_numpy()
        if principal_components == None : principal_components = self.principal_components
        
        preds = (self.transform(X, principal_components = principal_components)) @ self.loadings[:principal_components, :] 
        
        return preds
    
    def transform(self, X, principal_components = None): 
        
        """

        Transforms the X sample to the principal component space .

        Parameters
        ----------
        X : array_like
            Data used to extract the parameters and fit the model
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        result : array_like 
            returns X samples' scores.  

        """

        if isinstance(X, DataFrame) : X = X.to_numpy()            
        if principal_components == None : principal_components = self.principal_components
        
        if isnan(sum(X)) :
            result = zeros((X.shape[0], principal_components))
            X_nan = isnan(X)
            variables_missing_mask = unique(X_nan, axis = 0)
            
            for row_mask in variables_missing_mask:
                
                rows_indexes = where((X_nan == row_mask).all(axis = 1))                
                
                if sum(row_mask) == 0: 

                    result[rows_indexes, :] = X[rows_indexes, :] @ self.loadings[:principal_components, :].T 
                
                else:
                    
                    result[rows_indexes, :] = scores_with_missing_values(self.omega, self.loadings[:, ~row_mask], X[rows_indexes[0][:, None], ~row_mask], 
                                                                            LVs = principal_components, method = self.missing_values_method)
                 
                    
        else: result = X @ self.loadings[:principal_components, :].T 
        
        return result
    
    def score(self, X, principal_components = None): #return r²
        
        """
        
        Returns the coefficient of determination R^2 of the model.

        R² is defined as 1 - Variance(Error) / Variance(X) with Error = X - predictions(X)

        Parameters
        ----------
        X : array_like
            Data used to extract the parameters and fit the model
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        result : array_like 
            returns calculated r².        
        
        """

        if isinstance(X, DataFrame) : X = X.to_numpy()
        
        ErrorQ2 = X - self.predict(X, principal_components = principal_components)
        result = 1 - nanvar(ErrorQ2) / nanvar(X)
        return result

    def Hotellings_T2(self, X, principal_components = None):
        
        """
        Calculates the Hotelling's T² for the X samples.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        T2s : array_like 
            returns all calculated T²s for the X samples
        """

        if isinstance(X, DataFrame) : X = X.to_numpy()
            
        if principal_components == None : principal_components = self.principal_components # Unless specified, the number of PCs is the one in the trained model 
        
        scores_matrix = self.transform(X, principal_components = principal_components)
        
        T2s = sum((scores_matrix / self._training_scores_stds[:principal_components]) ** 2 , axis = 1)
        return T2s
    
    def T2_limit(self, alpha, principal_components = None):
        
        """
        Calculates the Hotelling's T² limit based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        t2_limit : array_like 
            returns the limit T² for the alpha based on the training dataset
        """

        if principal_components == None: principal_components = self.principal_components # Unless specified, the number of PCs is the one in the trained model 

        F_value = f.isf(1 - alpha , principal_components, self.training_scores.shape[0])
        t2_limit = ((principal_components * (self.training_scores.shape[0] ** 2 - 1)) / 
                    (self.training_scores.shape[0] * (self.training_scores.shape[0] - principal_components))) * \
                    F_value

        return t2_limit
    
    def _VIPs_calc(self, X, principal_components = None):
        
        if principal_components is None: principal_components = self.principal_components
              
        if isinstance(X, DataFrame) : X = X.to_numpy()

        SSX = array(sum((nan_to_num(X) - nanmean(X, axis = 0)) ** 2))   
        
        for i in range(1, principal_components + 1):

            pred = self.predict(X, principal_components = i)
            res = nan_to_num(X) - pred
            SSX = append(SSX, (res - res.mean(axis = 0) ** 2).sum())
            
        SSXdiff = SSX[:-1] - array(SSX)[1:]

        VIPs = array(((self.loadings[:principal_components, :].T ** 2) @ SSXdiff * 
                          self.loadings.shape[1] / (SSX[0] - SSX[len(SSX) - 1])) ** 1 / 2 , ndmin = 2)                
        return VIPs
    
    def contributions_scores_ind(self, X, principal_components = None): #contribution of each individual point in X1
        
        """
        calculates the sample individual contributions to the scores.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        contributions : array_like 
            matrix of scores contributions for every X sample
        """

        if principal_components == None : principal_components = self.principal_components

        if isinstance(X, DataFrame) : X = X.to_numpy()
        
        scores = self.transform(X, principal_components = principal_components)
        scores = (scores / self._training_scores_stds[:principal_components]) ** 2 
        contributions = multiply(X, (scores @ self.loadings[:principal_components, :] ** 2) ** (1 / 2)) 

        return contributions
       
    def contributions_spe(self, X, principal_components = None):
        
        """
        calculates the individual sample individual contributions to the squared prediction error 
        of the X variables matrix reconstruction.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_contributions : array_like 
            matrix of SPE contributions for every X sample
        """
        
        if principal_components == None : principal_components = self.principal_components
        if isinstance(X, DataFrame): X = X.to_numpy()

        error = nan_to_num(X) - self.predict(X)
        SPE_contributions = multiply(error ** 2 , where(error > 0, 1, -1))
        
        return SPE_contributions   

    def SPEs(self, X, principal_components = None):
        
        """
        Calculates the Squared prediction error for for the X matrix rebuild.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE : array_like 
            returns all calculated SPEs for the X samples
        """
        
        if principal_components == None : principal_components = self.principal_components
        if isinstance(X, DataFrame): X = X.to_numpy()

        error = nan_to_num(X) - self.predict(X)
        SPE = sum(error ** 2 , axis = 1)
        
        return SPE

    def SPE_limit(self, alpha, principal_components = None):

        """
        Calculates the SPE limit based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_limit : array_like 
            returns the limit SPE for the alpha based on the training dataset
        """

        if principal_components == None: principal_components = self.principal_components # Unless specified, the number of PCs is the one in the trained model 
        chi2_df = 2 * self._chi2_params[principal_components - 1][0] ** 2 / self._chi2_params[principal_components - 1][1]
        chi2_val = chi2.isf(1 - alpha, chi2_df)
        SPE_limit = self._chi2_params[principal_components - 1][1] / (2 * self._chi2_params[principal_components - 1][0])  * chi2_val
        
        return SPE_limit